[Version = "1.0.0"]
section Sitefinity.Insight;

[DataSource.Kind="Sitefinity.Insight", Publish="Sitefinity.Insight.Publish"]
shared Sitefinity.Insight.Contents = (optional NumberOfDaysToRetrieve as number) =>
    let
        _numOfDays = if(NumberOfDaysToRetrieve <> null) then NumberOfDaysToRetrieve else 1100,

        accessKey = Extension.CurrentCredential()[Key],

        apiServerUrl = Insight.ParseApiServerUrl(accessKey),

        token = Insight.GetAccessToken(apiServerUrl, accessKey),

        enabled = Insight.IsPowerBiEnabled(token, apiServerUrl),

        checkForPermission = if enabled then token else error "The Power BI Connector is available to active Sitefinity Insight premium subscriptions. Please contact your Progress Software sales representative or sales@progress.com.",

        dcs = Insight.GetDatacenters(checkForPermission, apiServerUrl),

        apiKeys = Table.SelectColumns(dcs, {"ApiKey"}),
        
        base = #table(type table [ Name = Text.Type, Data = Table.Type, IsLeaf=Logical.Type ], {
            { "Datacenters", dcs, true },
            { "Periods", Insight.Periods(), true },
            { "VisitorsDaily", Insight.VisitorsDaily(token, apiServerUrl, apiKeys, _numOfDays), true },
            { "VisitorsPerPeriod", Insight.VisitorsPerPeriod(token, apiServerUrl, apiKeys), true },
            { "Personas", Insight.Personas(token, apiServerUrl, apiKeys), true },
            { "PersonasDaily", Insight.PersonasDaily(token, apiServerUrl, apiKeys, _numOfDays), true },
            { "PersonasPerPeriod", Insight.PersonasPerPeriod(token, apiServerUrl, apiKeys), true },
            { "Conversions", Insight.Conversions(token, apiServerUrl, apiKeys), true },
            { "ConversionsDaily", Insight.ConversionsDaily(token, apiServerUrl, apiKeys, _numOfDays), true },
            { "ConversionsPerPeriod", Insight.ConversionsPerPeriod(token, apiServerUrl, apiKeys), true },
            { "ConversionsByPersona", Insight.ConversionsByPersona(token, apiServerUrl, apiKeys), true },
            { "Touchpoints", Insight.Touchpoints(token, apiServerUrl, apiKeys), true },
            { "TouchpointsDaily", Insight.TouchpointDaily(token, apiServerUrl, apiKeys, _numOfDays), true },
            { "TouchpointsPerPeriod", Insight.TouchpointsPerPeriod(token, apiServerUrl, apiKeys), true },
            { "TouchpointHitsByPersona", Insight.TouchpointsByPersona(token, apiServerUrl, apiKeys), true },
            { "Uplift", Insight.Uplift(token, apiServerUrl, apiKeys), true },
            { "Leads", Insight.LeadStagesFlat(token, apiServerUrl, apiKeys), true },
            { "LeadsDaily", Insight.LeadStagesDaily(token, apiServerUrl, apiKeys, _numOfDays), true }
        }),

        nav = CreateNavTable(base)
    in
        nav;


// Data Source Kind description
Sitefinity.Insight = [
    TestConnection = (dataSourcePath) => { "Sitefinity.Insight.Contents" },
    Authentication = [
         Key = [
            KeyLabel = "Access Key (issued by insight.sitefinity.com)",
            Label = "Access Key"
         ]
    ],
    Label = Extension.LoadString("DataSourceLabel")
];

// Data Source UI publishing description
Sitefinity.Insight.Publish = [
    Beta = false,
    Category = "Online Services",
    ButtonText = { Extension.LoadString("ButtonTitle"), Extension.LoadString("ButtonHelp") },
    LearnMoreUrl = "https://www.progress.com/sitefinity-cms/insight",
    SourceImage = Sitefinity.Icons,
    SourceTypeImage = Sitefinity.Icons
];

Sitefinity.Icons = [
    Icon16 = { Extension.Contents("ProgressLogo16.png"), Extension.Contents("ProgressLogo20.png"), Extension.Contents("ProgressLogo24.png"), Extension.Contents("ProgressLogo32.png") },
    Icon32 = { Extension.Contents("ProgressLogo32.png"), Extension.Contents("ProgressLogo40.png"), Extension.Contents("ProgressLogo48.png"), Extension.Contents("ProgressLogo64.png") }
];

// 
// Load common library functions
// 
// TEMPORARY WORKAROUND until we're able to reference other M modules
Extension.LoadFunction = (name as text) =>
    let
        binary = Extension.Contents(name),
        asText = Text.FromBinary(binary)
    in
        Expression.Evaluate(asText, #shared);

Table.ToNavigationTable = Extension.LoadFunction("Table.ToNavigationTable.pqm");
CreateNavTable = (base as table) as table =>
    let
        withItemKind = Table.AddColumn(base, "ItemKind", each "Table", type text),
        withItemName = Table.AddColumn(withItemKind, "ItemName", each "Table", type text),
        navTable = Table.ToNavigationTable(withItemName, {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        navTable;

Insight.CalculateImprovement = (current, prev) => if prev = 0 then null else (current-prev)/prev;

Insight.ParseApiServerUrl = Extension.LoadFunction("Insight.ParseApiServerUrl.pqm");
Insight.GetAccessToken = Extension.LoadFunction("Insight.GetAccessToken.pqm");
Insight.Get = Extension.LoadFunction("Insight.Get.pqm");
Insight.GetByApiKey = (token, apiServerUrl, endpoint, apiKey, headers) => 
 let
     combinedHeaders = Record.Combine({headers, [#"x-dataintelligence-datacenterkey"=apiKey]}),
     res = Insight.Get(token, apiServerUrl, endpoint, combinedHeaders)
 in 
     res;

Insight.GetStatistics_AllDCs = (token, apiServerUrl, apiKeys, endpoint, numOfDays) => 
    let 
        fromText = DateTimeZone.ToText(Date.AddDays(DateTimeZone.UtcNow(), -numOfDays), "yyyy-MM-dd"),
        toText = DateTimeZone.ToText(DateTimeZone.UtcNow(), "yyyy-MM-dd"),
        headers = [#"x-dataintelligence-fromdate"=fromText,
                   #"x-dataintelligence-todate"=toText],

        stats = Table.AddColumn(apiKeys, "PerDC", each Insight.GetByApiKey(token, apiServerUrl, endpoint, [ApiKey], headers)),
        forAllDCs = Table.ExpandTableColumn(stats, "PerDC", {"Column1"}, {"Column1"}),
        filtered = Table.SelectRows(forAllDCs, each [Column1] <> null)       
    in
        filtered;

Insight.GetStatistics_AllDCs_V3= (token, apiServerUrl, apiKeys, endpoint, numOfDays) => 
    let 
        fromText = DateTimeZone.ToText(Date.AddDays(DateTimeZone.UtcNow(), -numOfDays), "yyyy-MM-dd"),
        toText = DateTimeZone.ToText(DateTimeZone.UtcNow(), "yyyy-MM-dd"),
        endpointWithParams = Text.Combine({endpoint, "?FromDate=", fromText, "&ToDate=", toText}),
        
        stats = Table.AddColumn(apiKeys, "PerDC", each Insight.Get(token, apiServerUrl, Text.Replace(endpointWithParams, "{apiKey}", [ApiKey]), [])),
        forAllDCs = Table.ExpandTableColumn(stats, "PerDC", {"Column1"}, {"Column1"}),
        filtered = Table.SelectRows(forAllDCs, each [Column1] <> null)       
    in
        filtered;

Insight.GetPeriodStatistics_AllDCs_V3 = (token, apiServerUrl, apiKeys, endpoint) => 
    let 
        stats = Table.AddColumn(apiKeys, "PerDC", each Insight.Get(token, apiServerUrl, Text.Replace(endpoint, "{apiKey}", [ApiKey]), [])),
        forAllDCs = Table.ExpandTableColumn(stats, "PerDC", {"Column1"}, {"Column1"}),
        filtered = Table.SelectRows(forAllDCs, each [Column1] <> null)       
    in
        filtered;

Insight.Get_AllDCs = (token, apiServerUrl, apiKeys, endpoint) => 
    let 
        convs = Table.AddColumn(apiKeys, "PerDC", each Insight.GetByApiKey(token, apiServerUrl, endpoint, [ApiKey], [])),
        forAllDCs = Table.ExpandTableColumn(convs, "PerDC", {"Column1"}, {"Column1"}),
        filtered = Table.SelectRows(forAllDCs, each [Column1] <> null)       
    in
        filtered;

Insight.Periods = () as table =>
let
    tbl = #table(type table [ PeriodId = Text.Type, Name = Text.Type ], {
        {"Yesterday",  "Yesterday"},
        {"Last7Days",  "Last 7 days"},
        {"Last30Days",  "Last 30 days"},
        {"Last90Days",  "Last 3 months"},
        {"Last180Days",  "Last 6 months"}
    }),
    keyed = Table.AddKey(tbl, {"PeriodId"}, true)
in
    keyed;

// ------------------------------------ DATACENTERS ---------------------------------------------------//
Insight.GetDatacenters = (token as text, apiServerUrl as text) =>
let 
    Tbl = Insight.Get(token, apiServerUrl, "admin/v1/users/mydatacenters", []),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"ApiKey", "Name", "Description", "PersonasCount", "GoalsCount", "CreatedOn", "LeadsCount", "TouchpointsCount", "PersonalizationReportsCount", "ABTestsCount", "TotalContacts"}, 
        {"ApiKey", "Name", "Description", "PersonasCount", "ConversionsCount", "CreatedOn", "LeadsCount", "TouchpointsCount", "PersonalizationReportsCount", "ABTestsCount", "TotalVisitors"}),
    newTypes = Table.TransformColumnTypes(Expanded, {
        {"ApiKey", Guid.Type}, {"PersonasCount", Int32.Type}, {"ConversionsCount", Int32.Type}, {"LeadsCount", Int32.Type},  {"TotalVisitors", Int32.Type},
        {"TouchpointsCount", Int32.Type}, {"PersonalizationReportsCount", Int32.Type}, {"ABTestsCount", Int32.Type}, {"CreatedOn", type datetimezone}}),
    WithPK = Table.AddKey(newTypes, {"ApiKey"}, true)
in
    WithPK;

Insight.IsPowerBiEnabled = (token as text, apiServerUrl as text) =>
let 
    Tbl = Insight.Get(token, apiServerUrl, "/admin/v1/users/myaccounts", []),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", {"PremiumFeatures"}, {"PremiumFeatures"}),
    ExpList = Table.ExpandListColumn(Expanded, "PremiumFeatures"),
    AllFeatures = Table.ExpandRecordColumn(ExpList, "PremiumFeatures", {"Feature"}, {"Feature"}),
    IsPowerBiEnabled = Table.Contains(AllFeatures, [Feature = "GoogleDataStudio"])
in
    IsPowerBiEnabled;

Insight.VisitorsDaily = (token, apiServerUrl, apiKeys, numOfDays) => 
let 
    Tbl = Insight.GetStatistics_AllDCs(token, apiServerUrl, apiKeys, "analytics/v1/statistics/data-center/daily", numOfDays),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"StatisticsDate", "UniqueActiveVisitors", "UniqueNewContacts"}, 
        {"Date", "ActiveVisitors", "NewContacts"}),
    ChangeType = Table.TransformColumnTypes(Expanded,{{"ApiKey", Guid.Type}, {"Date", type datetimezone }, {"ActiveVisitors", Int32.Type}, {"NewContacts", Int32.Type}}),
    WithPK = Table.AddKey(ChangeType, {"ApiKey", "Date"}, true)
in
    WithPK;

Insight.VisitorsPerPeriod = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/statistics/period-metrics"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"StatisticsPeriod", "NewContacts_AllTime", "NewContacts_CurrentPeriod", "NewContacts_PreviousPeriod", "UniqueActiveVisitors_AllTime", "UniqueActiveVisitors_CurrentPeriod", 
        "UniqueActiveVisitors_PreviousPeriod", "UniqueVisitorsInAllPersonas_AllTime", "UniqueVisitorsInAllPersonas_CurrentPeriod"}, 
        {"PeriodId", "NewContacts_AllTime", "NewContacts_CurrentPeriod", "NewContacts_PreviousPeriod", "UniqueActiveVisitors_AllTime", "UniqueActiveVisitors_CurrentPeriod", 
        "UniqueActiveVisitors_PreviousPeriod", "UniqueVisitorsInAllPersonas_AllTime", "UniqueVisitorsInAllPersonas_CurrentPeriod"}), 
    NewContactsImprovement = Table.AddColumn(Expanded, "NewContacts_Improvement", each Insight.CalculateImprovement([NewContacts_CurrentPeriod], [NewContacts_PreviousPeriod])),
    VisitorsImprovement = Table.AddColumn(NewContactsImprovement, "UniqueActiveVisitors_Improvement", each Insight.CalculateImprovement([UniqueActiveVisitors_CurrentPeriod], [UniqueActiveVisitors_PreviousPeriod])),
    ChangeTypes = Table.TransformColumnTypes(VisitorsImprovement,{{"PeriodId", Text.Type},
        {"NewContacts_AllTime", Int32.Type},
        {"NewContacts_CurrentPeriod", Int32.Type}, 
        {"NewContacts_PreviousPeriod",  Int32.Type},
        {"NewContacts_Improvement", Percentage.Type},
        {"UniqueActiveVisitors_AllTime", Int32.Type}, 
        {"UniqueActiveVisitors_CurrentPeriod", Int32.Type},
        {"UniqueActiveVisitors_PreviousPeriod", Int32.Type},
        {"UniqueActiveVisitors_Improvement", Percentage.Type},
        {"UniqueVisitorsInAllPersonas_AllTime", Int32.Type},
        {"UniqueVisitorsInAllPersonas_CurrentPeriod", Int32.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"ApiKey", "PeriodId"}, true)
in
    WithPK;

// ------------------------------------ PERSONAS ---------------------------------------------------//
Insight.Personas = (token, apiServerUrl, apiKeys) => 
let
    Tbl = Insight.Get_AllDCs(token, apiServerUrl, apiKeys,  "analytics/v1/scorings/personas"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"Id", "Name", "DistinctiveCharacteristic", "State", "Threshold", "CreatedOn", "ModifiedOn"}, 
        {"Id", "Name", "DistinctiveCharacteristic", "State", "Threshold", "CreatedOn", "ModifiedOn"}),
    CompIdCol = Table.AddColumn(Expanded, "PersonaId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    noIdCol = Table.RemoveColumns(CompIdCol, {"Id", "ApiKey"}),
    typed = Table.TransformColumnTypes(noIdCol, {{"State", Int32.Type}, {"Threshold", Int32.Type}, {"CreatedOn", type datetimezone}, {"ModifiedOn", type datetimezone}}),
    WithPK = Table.AddKey(typed, {"PersonaId"}, true)
in
    WithPK;

Insight.PersonasDaily = (token, apiServerUrl, apiKeys, numOfDays) => 
let 
    Tbl =Insight.GetStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/personas/all/daily-statistics", numOfDays),
    Expanded  = Table.ExpandRecordColumn(Tbl, "Column1",
        {"PersonaId", "StatisticsDate", "VisitorsCount", "ContactsCount"}, 
        {"Id", "Date", "DailyVisitorsAboveThreshold", "DailyContactsAboveThreshold"}),
    changeType = Table.TransformColumnTypes(Expanded,{ {"Date", type datetimezone}, {"ApiKey", Guid.Type}, {"DailyVisitorsAboveThreshold", Int32.Type}, {"DailyContactsAboveThreshold", Int32.Type}}),
    CompIdCol = Table.AddColumn(changeType, "PersonaId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    noIdCol = Table.RemoveColumns(CompIdCol, {"Id"}),
    WithPK = Table.AddKey(noIdCol, {"PersonaId", "Date"}, true)
in
    WithPK;

Insight.PersonasPerPeriod = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/personas/all/period-metrics"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"PersonaId", "StatisticsPeriod", "VisitorsInPersona_CurrentPeriod", "VisitorsInPersona_PreviousPeriod", "VisitorsInPersona_AllTime", "ContactsInPersona_CurrentPeriod", "ContactsInPersona_PreviousPeriod", 
        "ContactsInPersona_AllTime", "VisitorsInPersonaToAllVisitorsRatio_CurrentPeriod", "VisitorsInPersonaToAllVisitorsRatio_PreviousPeriod", "VisitorsInPersonaToAllVisitorsRatio_AllTime", 
        "VisitorsInPersonaToVisitorsInAllPersonas_CurrentPeriod", "VisitorsInPersonaToVisitorsInAllPersonas_PreviousPeriod", "VisitorsInPersonaToVisitorsInAllPersonas_AllTime"}, 
        {"Id", "PeriodId", "VisitorsPassedThreshold_CurrentPeriod", "VisitorsPassedThreshold_PreviousPeriod", "VisitorsPassedThreshold_AllTime", "ContactsPassedThreshold_CurrentPeriod", "ContactsPassedThreshold_PreviousPeriod", 
        "ContactsPassedThreshold_AllTime", "ShareOfAllVisitors_CurrentPeriod", "ShareOfAllVisitors_PreviousPeriod", "ShareOfAllVisitors_AllTime", 
        "ShareOfAllPersonas_CurrentPeriod", "ShareOfAllPersonas_PreviousPeriod", "ShareOfAllPersonas_AllTime"}), 
    ConversionId = Table.AddColumn(Expanded, "PersonaId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    VisitorsPassedThreshold_Improvement = Table.AddColumn(ConversionId, "VisitorsPassedThreshold_Improvement", each Insight.CalculateImprovement([VisitorsPassedThreshold_CurrentPeriod], [VisitorsPassedThreshold_PreviousPeriod])),
    ContactsPassedThreshold_Improvement = Table.AddColumn(VisitorsPassedThreshold_Improvement, "ContactsPassedThreshold_Improvement", each Insight.CalculateImprovement([ContactsPassedThreshold_CurrentPeriod], [ContactsPassedThreshold_PreviousPeriod])),
    ShareOfAllVisitors_Improvement = Table.AddColumn(ContactsPassedThreshold_Improvement, "ShareOfAllVisitors_Improvement", each Insight.CalculateImprovement([ShareOfAllVisitors_CurrentPeriod], [ShareOfAllVisitors_PreviousPeriod])),
    ShareOfAllPersonas_Improvement = Table.AddColumn(ShareOfAllVisitors_Improvement, "ShareOfAllPersonas_Improvement", each Insight.CalculateImprovement([ShareOfAllPersonas_CurrentPeriod], [ShareOfAllPersonas_PreviousPeriod])),
    RemoveIdAndApiKey = Table.RemoveColumns(ShareOfAllPersonas_Improvement, {"Id"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveIdAndApiKey,{{"PersonaId", type text}, {"PeriodId", Text.Type}, {"ApiKey", Guid.Type},
        {"VisitorsPassedThreshold_CurrentPeriod", Int32.Type},
        {"VisitorsPassedThreshold_PreviousPeriod", Int32.Type}, 
        {"VisitorsPassedThreshold_AllTime",  Int32.Type},
        {"VisitorsPassedThreshold_Improvement",  Percentage.Type},
        
        {"ContactsPassedThreshold_CurrentPeriod", Int32.Type}, 
        {"ContactsPassedThreshold_PreviousPeriod", Int32.Type},
        {"ContactsPassedThreshold_AllTime", Int32.Type},
        {"ContactsPassedThreshold_Improvement",  Percentage.Type},
        
        {"ShareOfAllVisitors_CurrentPeriod", Number.Type},
        {"ShareOfAllVisitors_PreviousPeriod", Number.Type}, 
        {"ShareOfAllVisitors_AllTime", Number.Type},
        {"ShareOfAllVisitors_Improvement",  Percentage.Type},
        
        {"ShareOfAllPersonas_CurrentPeriod", Number.Type}, 
        {"ShareOfAllPersonas_PreviousPeriod", Number.Type}, 
        {"ShareOfAllPersonas_AllTime", Number.Type},
        {"ShareOfAllPersonas_Improvement",  Percentage.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"PersonaId", "PeriodId"}, true)
in
    WithPK;

// ------------------------------------ CONVERSIONS ---------------------------------------------------//
Insight.Conversions = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.Get_AllDCs(token, apiServerUrl,  apiKeys, "analytics/v1/conversions"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"Id", "Name", "State", "InputDays", "CreatedOn", "StartDate", "ModifiedOn"}, 
        {"Id", "Name", "State", "ConversionWindow", "CreatedOn", "StartDate", "ModifiedOn"}),
    CompIdCol = Table.AddColumn(Expanded, "ConversionId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    RemoveIdAndApiKey = Table.RemoveColumns(CompIdCol, {"Id", "ApiKey"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveIdAndApiKey,{{"ConversionId", type text}, {"Name", type text}, {"State", Int32.Type}, {"ConversionWindow", Int32.Type}, {"CreatedOn", type datetime}, {"StartDate", type datetimezone}, {"ModifiedOn", type datetimezone}}),
    WithPK = Table.AddKey(ChangeTypes, {"ConversionId"}, true)
in
    WithPK;

Insight.ConversionsDaily = (token, apiServerUrl, apiKeys, numOfDays) => 
let 
    Tbl = Insight.GetStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/conversions/all/daily-statistics", numOfDays),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"ConversionId", "StatisticsDate", "UniqueConversions"}, 
        {"Id", "Date", "UniqueConversions"}),
    ConversionId = Table.AddColumn(Expanded, "ConversionId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    RemoveId = Table.RemoveColumns(ConversionId, {"Id"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveId,{{"ApiKey", Guid.Type}, {"Date", type datetimezone}, {"ConversionId", type text}, {"UniqueConversions", Int32.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"ConversionId", "Date"}, true)
in
    WithPK;

Insight.ConversionsPerPeriod = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/conversions/all/period-metrics"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"ConversionId", "StatisticsPeriod", "VisitorsInSegment_CurrentPeriod", "VisitorsInSegment_PreviousPeriod", "VisitorsInSegment_AllTime", "TotalConversions_CurrentPeriod", "TotalConversions_PreviousPeriod", "TotalConversions_AllTime", "VisitorsConverted_CurrentPeriod", "VisitorsConverted_PreviousPeriod", "VisitorsConverted_AllTime", "ContactsConverted_CurrentPeriod", "ContactsConverted_PreviousPeriod", "ContactsConverted_AllTime", "NewContacts_CurrentPeriod", "NewContacts_PreviousPeriod", "NewContacts_AllTime", "ConversionRate"}, 
        {"Id", "PeriodId", "VisitorsInSegment_CurrentPeriod", "VisitorsInSegment_PreviousPeriod", "VisitorsInSegment_AllTime", "TotalConversions_CurrentPeriod", "TotalConversions_PreviousPeriod", "TotalConversions_AllTime", "VisitorsConverted_CurrentPeriod", "VisitorsConverted_PreviousPeriod", "VisitorsConverted_AllTime", "ContactsConverted_CurrentPeriod", "ContactsConverted_PreviousPeriod", "ContactsConverted_AllTime", "NewContacts_CurrentPeriod", "NewContacts_PreviousPeriod", "NewContacts_AllTime", "ConversionRate"}),
    ConversionId = Table.AddColumn(Expanded, "ConversionId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    VisitorsInSegment_Improvement = Table.AddColumn(ConversionId, "VisitorsInSegment_Improvement", each Insight.CalculateImprovement([VisitorsInSegment_CurrentPeriod], [VisitorsInSegment_PreviousPeriod])),
    TotalConversions_Improvement = Table.AddColumn(VisitorsInSegment_Improvement, "TotalConversions_Improvement", each Insight.CalculateImprovement([TotalConversions_CurrentPeriod], [TotalConversions_PreviousPeriod])),
    VisitorsConverted_Improvement = Table.AddColumn(TotalConversions_Improvement, "VisitorsConverted_Improvement", each Insight.CalculateImprovement([VisitorsConverted_CurrentPeriod], [VisitorsConverted_PreviousPeriod])),
    ContactsConverted_Improvement = Table.AddColumn(VisitorsConverted_Improvement, "ContactsConverted_Improvement", each Insight.CalculateImprovement([ContactsConverted_CurrentPeriod], [ContactsConverted_PreviousPeriod])),
    NewContacts_Improvement = Table.AddColumn(ContactsConverted_Improvement, "NewContacts_Improvement", each Insight.CalculateImprovement([NewContacts_CurrentPeriod], [NewContacts_PreviousPeriod])),
    RemoveId = Table.RemoveColumns(NewContacts_Improvement, {"Id"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveId,{{"ConversionId", type text}, {"PeriodId", Text.Type}, {"ApiKey", Guid.Type},
        {"VisitorsInSegment_CurrentPeriod", Int32.Type},
        {"VisitorsInSegment_PreviousPeriod", Int32.Type},
        {"VisitorsInSegment_AllTime", Int32.Type},
        {"VisitorsInSegment_Improvement", Percentage.Type},
        
        {"TotalConversions_CurrentPeriod", Int32.Type},
        {"TotalConversions_PreviousPeriod", Int32.Type},
        {"TotalConversions_AllTime", Int32.Type},
        {"TotalConversions_Improvement", Percentage.Type},
        
        {"VisitorsConverted_CurrentPeriod", Int32.Type},
        {"VisitorsConverted_PreviousPeriod", Int32.Type},
        {"VisitorsConverted_AllTime", Int32.Type},
        {"VisitorsConverted_Improvement", Percentage.Type},
        
        {"ContactsConverted_CurrentPeriod", Int32.Type},
        {"ContactsConverted_PreviousPeriod", Int32.Type},
        {"ContactsConverted_AllTime", Int32.Type},
        {"ContactsConverted_Improvement", Percentage.Type},
        
        {"NewContacts_CurrentPeriod", Int32.Type},
        {"NewContacts_PreviousPeriod", Int32.Type},
        {"NewContacts_AllTime", Int32.Type},
        {"NewContacts_Improvement", Percentage.Type},

        {"ConversionRate", Number.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"ConversionId", "PeriodId"}, true)
in
    WithPK;

Insight.Uplift = (token, apiServerUrl, apiKeys) => 
    let 
        headers = [#"x-dataintelligence-filter"="{""TouchpointType"":[""Touchpoint"",""Content""]}"],
        endpoint = "/analytics/v1/attributions",

        Tbl = Table.AddColumn(apiKeys, "PerDC", each Insight.GetByApiKey(token, apiServerUrl, endpoint, [ApiKey], headers)),
        ForAllDCs = Table.ExpandTableColumn(Tbl, "PerDC", {"Column1"}, {"Column1"}),
        Filtered = Table.SelectRows(ForAllDCs, each [Column1] <> null)  ,
        Expanded = Table.ExpandRecordColumn(Filtered, "Column1", 
            {"TouchpointId", "ConversionId", "Attribution", "TouchpointHits", "TouchpointUniqueHits", "GoalCompletions", "ConversionRateTouchpoint", "ConversionRateAllContacts"}, 
            {"TpId", "ConvId", "Uplift", "TouchpointHits", "TouchpointUniqueHits", "Conversions", "ConversionRatePerTouchpoint", "ConversionRate"}),
        ConversionId = Table.AddColumn(Expanded, "ConversionId", each Text.Combine( {[ApiKey], Number.ToText([ConvId])}, ":")),
        TouchpointId = Table.AddColumn(ConversionId, "TouchpointId", each Text.Combine( {[ApiKey], Number.ToText([TpId])}, ":")),
        RemoveId = Table.RemoveColumns(TouchpointId, {"ConvId", "TpId"}),
        ChangeTypes = Table.TransformColumnTypes(RemoveId, {{"ConversionId", type text}, {"TouchpointId", type text}, {"ApiKey", Guid.Type},
            {"Uplift", Number.Type}, 
            {"TouchpointHits", Int32.Type},
            {"TouchpointUniqueHits", Int32.Type},
            {"Conversions", Int32.Type},
            {"ConversionRatePerTouchpoint", Number.Type},
            {"ConversionRate", Number.Type}}),
        WithPK = Table.AddKey(ChangeTypes, {"ConversionId", "TouchpointId"}, true)
    in
        WithPK;

Insight.ConversionsByPersona = (token, apiServerUrl, apiKeys) => 
    let 
        Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/personas/all/conversions/period-metrics"),
        Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
            {"PersonaId", "ConversionId", "TotalVisitorsCount", "StatisticsPeriod"}, 
            {"pId", "ConvId", "Conversions", "PeriodId"}),
        ConversionId = Table.AddColumn(Expanded, "ConversionId", each Text.Combine( {[ApiKey], Number.ToText([ConvId])}, ":")),
        PersonaId = Table.AddColumn(ConversionId, "PersonaId", each Text.Combine( {[ApiKey], Number.ToText([pId])}, ":")),
        RemoveCols = Table.RemoveColumns(PersonaId, {"ConvId", "pId"}),
        ChangeTypes = Table.TransformColumnTypes(RemoveCols, {{"ConversionId", type text}, {"PersonaId", type text}, {"ApiKey", Guid.Type}, {"Conversions", Int32.Type},  {"PeriodId", Text.Type}}),
        WithPK = Table.AddKey(ChangeTypes, {"ConversionId", "PersonaId", "PeriodId"}, true)
    in
        WithPK;

// ------------------------------------ Touchpoints ---------------------------------------------------//
Insight.Touchpoints = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.Get_AllDCs(token, apiServerUrl, apiKeys, "analytics/v1/touchpoints"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"Id", "Name", "State", "TotalHits", "AbsoluteHits", "CreatedOn", "ModifiedOn"}, 
        {"Id", "Name", "State", "TotalHits", "AbsoluteHits", "CreatedOn", "ModifiedOn"}),
    CompIdCol = Table.AddColumn(Expanded, "TouchpointId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    RemoveIdAndApiKey = Table.RemoveColumns(CompIdCol, {"Id", "ApiKey"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveIdAndApiKey,{{"Name", type text}, {"TotalHits", Int32.Type}, {"CreatedOn", type datetime}, {"ModifiedOn", type datetimezone}, {"AbsoluteHits", Int32.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"TouchpointId"}, true)
in
    WithPK;

Insight.TouchpointDaily = (token, apiServerUrl, apiKeys, numOfDays) => 
let 
    Tbl = Insight.GetStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/touchpoints/all/daily-statistics", numOfDays),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"TouchpointId", "StatisticsDate", "UniqueHits", "TotalHits"}, 
        {"Id", "Date", "UniqueHits", "TotalHits"}),
    CompIdCol = Table.AddColumn(Expanded, "TouchpointId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    RemoveId = Table.RemoveColumns(CompIdCol, {"Id"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveId,{{"Date", type datetimezone}, {"ApiKey", Guid.Type}, {"UniqueHits", Int32.Type}, {"TotalHits", Int32.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"TouchpointId", "Date"}, true)
in
    WithPK;

Insight.TouchpointsPerPeriod = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/touchpoints/all/period-metrics"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"TouchpointId", "StatisticsPeriod", "TotalHits_CurrentPeriod", "TotalHits_PreviousPeriod", "TotalHits_AllTime", 
         "UniqueHits_CurrentPeriod", "UniqueHits_PreviousPeriod", "UniqueHits_AllTime", "ContactsHits_CurrentPeriod",
        "ContactsHits_PreviousPeriod", "ContactsHits_AllTime"}, 
        {"Id", "PeriodId", "TotalHits_CurrentPeriod", "TotalHits_PreviousPeriod", "TotalHits_AllTime",
        "UniqueHits_CurrentPeriod", "UniqueHits_PreviousPeriod", "UniqueHits_AllTime", "ContactsHits_CurrentPeriod", 
        "ContactsHits_PreviousPeriod", "ContactsHits_AllTime"}),
    CompIdCol = Table.AddColumn(Expanded, "TouchpointId", each Text.Combine( {[ApiKey], Number.ToText([Id])}, ":")),
    TotalHits_Improvement = Table.AddColumn(CompIdCol, "TotalHits_Improvement", each Insight.CalculateImprovement([TotalHits_CurrentPeriod], [TotalHits_PreviousPeriod])),
    UniqueHits_Improvement = Table.AddColumn(TotalHits_Improvement, "UniqueHits_Improvement", each Insight.CalculateImprovement([UniqueHits_CurrentPeriod], [UniqueHits_PreviousPeriod])),
    ContactsHits_Improvement = Table.AddColumn(UniqueHits_Improvement, "ContactsHits_Improvement", each Insight.CalculateImprovement([ContactsHits_CurrentPeriod], [ContactsHits_PreviousPeriod])),

    RemoveId = Table.RemoveColumns(ContactsHits_Improvement, {"Id"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveId,{{"TouchpointId", type text}, {"PeriodId", Text.Type}, {"ApiKey", Guid.Type},
        {"TotalHits_CurrentPeriod", Int32.Type}, 
        {"TotalHits_PreviousPeriod", Int32.Type},  
        {"TotalHits_AllTime", Int32.Type}, 
        {"TotalHits_Improvement", Percentage.Type}, 
        
        {"UniqueHits_CurrentPeriod", Int32.Type},
        {"UniqueHits_PreviousPeriod", Int32.Type},
        {"UniqueHits_AllTime", Int32.Type},
        {"UniqueHits_Improvement", Percentage.Type}, 
        
        {"ContactsHits_CurrentPeriod", Int32.Type},
        {"ContactsHits_PreviousPeriod", Int32.Type},
        {"ContactsHits_AllTime", Int32.Type},
        {"ContactsHits_Improvement", Percentage.Type}}),
    WithPK = Table.AddKey(ChangeTypes, {"TouchpointId", "PeriodId"}, true)
in
    WithPK;

Insight.TouchpointsByPersona = (token, apiServerUrl, apiKeys) => 
    let 
        Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/personas/all/touchpoints/period-metrics"),
        Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
            {"PersonaId", "TouchpointId", "TotalHits", "StatisticsPeriod"}, 
            {"pId", "tId", "TotalHits", "PeriodId"}),
        TouchpointId = Table.AddColumn(Expanded, "TouchpointId", each Text.Combine( {[ApiKey], Number.ToText([tId])}, ":")),
        PersonaId = Table.AddColumn(TouchpointId, "PersonaId", each Text.Combine( {[ApiKey], Number.ToText([pId])}, ":")),
        RemoveCols = Table.RemoveColumns(PersonaId, {"tId", "pId"}),
        ChangeTypes = Table.TransformColumnTypes(RemoveCols, {{"TouchpointId", type text}, {"PersonaId", type text}, {"ApiKey", Guid.Type}, {"TotalHits", Int32.Type},  {"PeriodId", Text.Type}}),
        WithPK = Table.AddKey(ChangeTypes, {"TouchpointId", "PersonaId", "PeriodId"}, true)
    in
        WithPK;

// ------------------------------------ LEADS ---------------------------------------------------//
Insight.LeadStagesStats = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.GetPeriodStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/leads/all/period-metrics"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"LeadId", "LevelId", "VisitorsInLevel_CurrentPeriod", "ContactsInLevel_CurrentPeriod", "VisitorsPassedLevel_CurrentPeriod"}, 
        {"LId", "LvlId", "VisitorsInLevel", "ContactsInLevel", "VisitorsPassedLevel"} ),
    StageId = Table.AddColumn(Expanded, "StageId", each Text.Combine( {[ApiKey], Number.ToText([LId]),  Number.ToText([LvlId])}, ":")),
    RemoveIdAndApiKey = Table.RemoveColumns(StageId, {"ApiKey", "LvlId", "LId"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveIdAndApiKey,{{"StageId", type text},
        {"VisitorsInLevel", Int32.Type}, 
        {"ContactsInLevel", Int32.Type},  
        {"VisitorsPassedLevel", Int32.Type}})
in
    ChangeTypes;

Insight.LeadStagesFlat = (token, apiServerUrl, apiKeys) => 
let 
    Tbl = Insight.Get_AllDCs(token, apiServerUrl, apiKeys, "analytics/v1/scorings/leads"),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"Id", "Name", "Levels"}, 
        {"LId",  "LeadName", "Levels"}),
    LevelsExpanded = Table.ExpandListColumn(Expanded, "Levels"),
    LevelsExpandedRaws = Table.ExpandRecordColumn(LevelsExpanded, "Levels", {"Id", "Name", "Threshold", "LastThresholdPassedDate"}, {"LvlId", "Stage", "StageThreshold", "LastPassedOn"}),
    
    LevelId = Table.AddColumn(LevelsExpandedRaws, "StageId", each Text.Combine( {[ApiKey], Number.ToText([LId]),  Number.ToText([LvlId])}, ":")),
    
    RemoveId = Table.RemoveColumns(LevelId, {"LId", "LvlId", "ApiKey"}),
    ChangeTypes = Table.TransformColumnTypes(RemoveId, {{"LeadName", type text}, {"Stage", type text}, {"StageId", type text}, {"StageThreshold", Int32.Type}, {"LastPassedOn", type datetimezone}}),
    
    Stats = Insight.LeadStagesStats(token, apiServerUrl, apiKeys),

    JointedTables = Table.NestedJoin(ChangeTypes, {"StageId"}, Stats, {"StageId"}, "Stats", JoinKind.LeftOuter),
    ExpandedStats = Table.ExpandTableColumn(JointedTables, "Stats", {"VisitorsInLevel", "ContactsInLevel", "VisitorsPassedLevel"}, {"VisitorsInStage", "ContactsInStage", "VisitorsPassedStage"}),
    
    WithPK = Table.AddKey(ExpandedStats, {"StageId"}, true)
in 
   WithPK;

//TODO: decide if we need both Passed and In metrics 
Insight.LeadStagesDaily = (token, apiServerUrl, apiKeys, numOfDays) => 
let 
    Tbl = Insight.GetStatistics_AllDCs_V3(token, apiServerUrl, apiKeys, "/v3/data-centers/{apiKey}/scorings/leads/all/daily-statistics", numOfDays),
    Expanded = Table.ExpandRecordColumn(Tbl, "Column1", 
        {"ScoringId", "ScoringThresholdId", "DailyClientsPassedThreshold", "DailyKnownClientsPassedThreshold", "StatisticsDate"}, 
        {"LId", "LvlId", "Visitors", "Contacts", "Date"}),
    StageId = Table.AddColumn(Expanded, "StageId", each Text.Combine( {[ApiKey], Number.ToText([LId]),  Number.ToText([LvlId])}, ":")),
    ChangedType = Table.TransformColumnTypes(StageId,{ {"StageId", type text}, {"ApiKey", Guid.Type}, {"Visitors", Int32.Type}, {"Contacts", Int32.Type}, {"Date", type datetimezone}}),
    RemoveIds = Table.RemoveColumns(ChangedType, {"LId", "LvlId"}),
    WithPK = Table.AddKey(RemoveIds, {"Date", "StageId"}, true)
in
    WithPK;